/* packet-ut.c
 * Routines for Upper Tester Protocol dissection
 * Copyright 2015, AMB Consulting <alexandre.berge@amb-consulting.com>
 *
 * $Id$
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

//#ifdef HAVE_CONFIG_H
#include "config.h"
//#endif

#include <glib.h>

#include <epan/packet.h>
#include <epan/expert.h>
#include <epan/prefs.h>

#define UT_PROTOCOL_PORT                12345 

#define MIN_PDU_SIZE                    2 

#define UT_INITIALISE                   0x00 /* Refer to Draft ETSI TR 103 099 V1.5.0 (2014-12) Annex C */
#define UT_INITIALISE_RESULT            0x01
#define UT_CHANGEPOSITION               0x02
#define UT_CHANGEPOSITION_RESULT        0x03
#define UT_CHANGEPSEUDONYM              0x04
#define UT_CHANGEPSEUDONYM_RESULT       0x05
/* CAM */
#define UT_CHANGECURVATURE              0x30
#define UT_CHANGESPEED                  0x31
#define UT_SETACCELERATIONCONTROLSTATUS 0x32
#define UT_SETEXTERIORLIGHTSSTATUS      0x33
#define UT_CHANGEHEADING                0x34
#define UT_SETDRIVEDIRECTION            0x35
#define UT_CHANGEYAWRATE                0x36
#define UT_CAMEVENTINDICATION           0x23
#define UT_SETSTATIONTYPE               0x39
#define UT_SETVEHICLEROLE               0x3a
#define UT_SETEMBARKATIONSTATUS         0x3b
#define UT_SETPTACTIVATION              0x3c
#define UT_SETDANGEROUSGOODS            0x3d
#define UT_SETDANGEROUSGOODSEXT         0x3e
#define UT_SETLIGHTBARSIREN             0x3f
#define UT_CAMTRIGGER_RESULT            0x21
/* DENM */
#define UT_DENMTRIGGER                  0x10
#define UT_DENMTRIGGER_RESULT           0x11
#define UT_DENMEVENT                    0x12
#define UT_DENMEVENT_RESULT             0x13
#define UT_TERMINATE_DENMEVENT          0x14
#define UT_TERMINATE_DENMEVENT_RESULT   0x15
#define UT_DENMEVENTINDICATION          0x17
/* GN */
#define UT_GENERATEGEOUNICAST           0x50
#define UT_GENERATEGEOBROADCAST         0x51
#define UT_GENERATEGEOANYCAST           0x52
#define UT_GENERATEGEOSHB               0x53
#define UT_GENERATEGEOTSB               0x54
#define UT_GNEVENTINDICATION            0x55
#define UT_GNTRIGGER_RESULT             0x41
/* MAP/SPaT */
#define UT_MAPSPAT_TRIGGER_EVENT        0xA0
#define UT_MAPSPAT_TRIGGER_EVENT_RESULT 0xA1


/* Names table indexes */
#define IDX_RESULT_OK                   0x01
#define IDX_RESULT_KO                   0x00
#define IDX_SECUREDMODE_DISABLE         0x00
#define IDX_SECUREDMODE_ENABLE          0xFF

/* Messages length */
/* General */
#define L_UT_COMMAND                    1
#define L_UT_FLAGS                      1
#define L_UT_INITIALISE                 9
#define L_UT_INITIALISE_RESULT          2
#define L_UT_INITIALISE_RESULT_CODE     1
#define L_UT_CHANGEPOSITION             10
#define L_UT_CHANGEPOSITION_RESULT      2
#define L_UT_CHANGEPOSITION_RESULT_CODE 1
#define L_UT_CHANGEPSEUDONYM            1
#define L_UT_CHANGEPSEUDONYM_RESULT     2
#define L_UT_CHANGEPSEUDONYM_RESULT_CODE 1
/* CAM */
#define L_UT_CHANGECURVATURE            3
#define L_UT_CHANGESPEED                3
#define L_UT_SETACCELERATIONCONTROLSTATUS 2
#define L_UT_SETEXTERIORLIGHTSSTATUS    2
#define L_UT_CHANGEHEADING              3
#define L_UT_SETDRIVEDIRECTION          2
#define L_UT_CAM_TRIGGER_RESULT         2
#define L_UT_CHANGEYAWRATE              3
#define L_UT_SETSTATIONTYPE             2
#define L_UT_SETVEHICLEROLE             2
#define L_UT_SETEMBARKATIONSTATUS       2
#define L_UT_SETDANGEROUSGOODS          2
#define L_UT_SETDANGEROUSGOODSEXT       2
#define L_UT_SETLIGHTBARSIREN           2
#define L_UT_CAM_TRIGGER_RESULT_CODE    1
/* DENM */
#define L_UT_DENMM_TRIGGER_RESULT       8
#define L_UT_DENMM_TRIGGER_RESULT_CODE  1
#define L_UT_DENMM_EVENT_RESULT         8
#define L_UT_DENMM_EVENT_RESULT_CODE    1
/* GN */
#define L_UT_GNTRIGGER_RESULT           2
#define L_UT_GNTRIGGER_RESULT_CODE      1
/* MAP/SPaT */
#define L_UT_MAPSPAT_TRIGGER_EVENT       2
#define L_UT_MAPSPAT_TRIGGER_EVENT_RESULT 3
#define L_UT_MAPSPAT_TRIGGER_EVENT_RESULT_CODE 1
#define L_UT_MAPSPAT_TRIGGER_EVENT_RESULT_MSGC 1

/* Parameters length */
#define L_HASHEDID8                     8
#define L_DELTALATITUDE                 4
#define L_DELTALONGITUDE                4
#define L_DELTAELEVATION                1
#define L_CURVATURE                     2
#define L_SPEEDVARIATION                2
#define L_HEADING_DIRECTION             2
#define L_DRIVE_DIRECTION               1
#define L_YAWRATE                       2
#define L_STATIONTYPE                   1
#define L_VEHICLEROLE                   1
#define L_EMBARKATIONSTATUS             1
#define L_PTACTIVATIONTYPE              1
#define L_PTACTIVATIONDATALENGTH        1
#define L_DANGEROUSGOODS                1
#define L_STATIONID                     4
#define L_SEQNUM                        2
#define L_DETECTIONTIME                 6
#define L_VALIDITYDURATION              3
#define L_REPETITIONDURATION            3
#define L_INFOQUALITY                   1
#define L_CAUSE                         1
#define L_SUBCAUSE                      1
#define L_RELEVANCEDISTANCE             1
#define L_RELEVANCETRAFFICDIRECTION     1
#define L_TRANSMISSIONINTERVAL          2
#define L_REPETITIONINTERVAL            2
#define L_ALACARTELENGTH                1
#define L_CAMEVENTLENGTH                2
#define L_DENMEVENTLENGTH               2
#define L_SETPTACTIVATIONLENGTH         1
#define L_SHAPE                         1
#define L_LIFETIME	                    2
#define L_TRAFFICCLASS	                1
#define L_RESERVEDGEOBROADCAST	        3
#define L_LATITUDE	                    4
#define L_LONGITUDE	                    4
#define L_DISTANCEA	                    2
#define L_DISTANCEB	                    2
#define L_ANGLE	                        2
#define L_PAYLOADLENGTH	                2
#define L_NBHOPS    	                1
#define L_DSTGNADDR    	                8
#define L_MAPSPAT_TRIGGER_EVENT    	    1

/* Function declarations */
void proto_reg_handoff_ut(void);

/* Masks */

/* Initialize the protocol and registered fields */
static int proto_ut = -1;
static int hf_command = -1;

/* UT Initialise */
static int hf_initialize = -1;
static int hf_hashed_id8 = -1;
static int hf_initialize_result = -1;
static int hf_initialize_result_code = -1;

/* UT Change Position */
static int hf_change_position = -1;
static int hf_change_position_result = -1;
static int hf_change_position_result_code = -1;

/* UT Change Pseudonym */
static int hf_change_pseudonym = -1;
static int hf_change_pseudonym_result = -1;
static int hf_change_pseudonym_result_code = -1;

/* UT Change Curvature */
static int hf_change_curvature = -1;

/* UT Change Speed */
static int hf_change_speed = -1;

/* UT Set Acceleration Control Status */
static int hf_set_acceleration_control_status = -1;
static int hf_set_acceleration_control_status_flags = -1;
static int hf_set_acceleration_control_status_flags_B_bit = -1;
static int hf_set_acceleration_control_status_flags_G_bit = -1;
static int hf_set_acceleration_control_status_flags_E_bit = -1;
static int hf_set_acceleration_control_status_flags_C_bit = -1;
static int hf_set_acceleration_control_status_flags_A_bit = -1;
static int hf_set_acceleration_control_status_flags_CC_bit = -1;
static int hf_set_acceleration_control_status_flags_L_bit = -1;

/* UT Set Exterior Lights Status */
static int hf_set_exterior_lights_status = -1;
static int hf_set_exterior_lights_status_flags = -1;
static int hf_set_exterior_lights_status_flags_LB_bit = -1;
static int hf_set_exterior_lights_status_flags_HB_bit = -1;
static int hf_set_exterior_lights_status_flags_LT_bit = -1;
static int hf_set_exterior_lights_status_flags_RT_bit = -1;
static int hf_set_exterior_lights_status_flags_D_bit = -1;
static int hf_set_exterior_lights_status_flags_R_bit = -1;
static int hf_set_exterior_lights_status_flags_F_bit = -1;
static int hf_set_exterior_lights_status_flags_P_bit = -1;

/* UT Change Heading */
static int hf_change_heading = -1;

/* UT Set Drive Direction */
static int hf_set_drive_direction = -1;

/* UT Change YawRate */
static int hf_change_yawrate = -1;

/* UT CAM Event Indication */
static int hf_cam_event_indication = -1;
static int hf_cam_event_indication_length = -1;
static int hf_cam_event_indication_payload = -1;

/* UT Set Station Type */
static int hf_set_station_type = -1;

/* UT Set Vehicle Role */
static int hf_set_vehicle_role = -1;

/* UT Set Embarkation Status */
static int hf_set_embarkation_status = -1;

/* UT Set Pt Activation */
static int hf_set_pt_activation = -1;
static int hf_set_pt_activation_length = -1;
static int hf_set_pt_activation_payload = -1;

/* UT Set Dangerous Goods */
static int hf_set_dangerous_goods = -1;
static int hf_set_dangerous_goods_ext = -1;

/* UT Set Light Bar Siren */
static int hf_set_light_bar_siren = -1;
static int hf_set_light_bar_siren_flags = -1;
static int hf_set_light_bar_siren_flags_LB_bit = -1;
static int hf_set_light_bar_siren_flags_S_bit = -1;

/* UT CAM Trigger */
static int hf_cam_trigger_result = -1;
static int hf_cam_trigger_result_code = -1;

/* UT DENM Trigger */
static int hf_denm_trigger = -1;
static int hf_denm_trigger_result = -1;
static int hf_denm_trigger_result_code = -1;
static int hf_denm_trigger_flags = -1;
static int hf_denm_trigger_flags_V_bit = -1;
static int hf_denm_trigger_flags_R_bit = -1;
static int hf_denm_trigger_flags_K_bit = -1;

/* UT DENM Event */
static int hf_denm_event = -1;
static int hf_denm_event_result = -1;
static int hf_denm_event_result_code = -1;
static int hf_denm_event_flags = -1;
static int hf_denm_event_flags_V_bit = -1;
static int hf_denm_event_flags_S_bit = -1;
static int hf_denm_event_flags_D_bit = -1;
static int hf_denm_event_flags_T_bit = -1;
static int hf_denm_event_flags_C_bit = -1;
static int hf_denm_event_flags_K_bit = -1;
static int hf_denm_event_flags_I_bit = -1;
static int hf_denm_event_flags_X_bit = -1;

/* UT DENM Terminate Event */
static int hf_terminate_denm_event = -1;
static int hf_terminate_denm_event_result = -1;
static int hf_denm_terminate_event_result_code = -1;

/* UT DENM Event Indication */
static int hf_denm_event_indication = -1;
static int hf_denm_event_indication_length = -1;
static int hf_denm_event_indication_payload = -1;

/* UT GN Generate Unicast */
static int hf_gn_geounicast = -1;

/* UT GN Generate Broadcast */
static int hf_gn_geobroadcast = -1;

/* UT GN Generate Antcast */
static int hf_gn_geoanycast = -1;

/* UT GN Generate SHB */
static int hf_gn_geoshb = -1;

/* UT GN Generate TSB */
static int hf_gn_geotsb = -1;

/* UT GN Event Indication */
static int hf_gn_event_indication = -1;
static int hf_gn_event_indication_length = -1;
static int hf_gn_event_indication_payload = -1;

/* UT GN Trigger Result */
static int hf_gntrigger_result = -1;
static int hf_gntrigger_result_code = -1;

/* UT MAP/SPaT Trigger Event */
static int hf_mapspat_trigger_event = -1;

/* UT MAP/SPaT Trigger Result */
static int hf_mapspat_trigger_event_result = -1;
static int hf_mapspat_trigger_event_result_code = -1;
static int hf_mapspat_trigger_event_result_msgc = -1;

/* CAM Parameters */
static int hf_delta_latitude = -1;
static int hf_delta_longitude = -1;
static int hf_delta_elevation = -1;
static int hf_speed_variation = -1;
static int hf_heading_direction = -1;
static int hf_drive_direction = -1;
static int hf_yawrate = -1;
static int hf_station_type = -1;
static int hf_vehicle_role = -1;
static int hf_embarkation_status = -1;
static int hf_dangerous_goods = -1;
static int hf_dangerous_goods_ext = -1;
/* DENM Parameters */
static int hf_station_id = -1;
static int hf_sequence_number = -1;
static int hf_detection_time = -1;
static int hf_validity_duration = -1;
static int hf_repetition_duration = -1;
static int hf_info_quality = -1;
static int hf_cause = -1;
static int hf_subcause = -1;
static int hf_relevance_distance = -1;
static int hf_relevance_traffic_direction = -1;
static int hf_transmission_interval = -1;
static int hf_repetition_interval = -1;
static int hf_alacarte_length = -1;
static int hf_alacarte = -1;
/* GN Parameters */
static int hf_gn_address = -1;
static int hf_shape = -1;
static int hf_lifetime = -1;
static int hf_trafficclass = -1;
static int hf_reservedgeobroadcast = -1;
static int hf_latitude = -1;
static int hf_longitude = -1;
static int hf_distancea = -1;
static int hf_distanceb = -1;
static int hf_angle = -1;
static int hf_payloadlength = -1;
static int hf_payload = -1;
static int hf_nbhops = -1;
static int hf_dstgnaddr = -1;
/* MAP/SPaT Parameters */
static int hf_event = -1;

/* Initialize the subtree pointers */
static gint ett_ut = -1;
static gint ett_ut_command = -1;
static gint ett_ut_flags = -1;
static gint ett_ut_data = -1;

static const value_string ut_command_names[] = {
    { UT_INITIALISE,                    "Initialise" },
    { UT_INITIALISE_RESULT,             "Initialize Result" },
    { UT_CHANGEPOSITION,                "Change Position" },
    { UT_CHANGEPOSITION_RESULT,         "Change Position Result" },
    { UT_CHANGEPSEUDONYM,               "Change Pseudonym" },
    { UT_CHANGEPSEUDONYM_RESULT,        "Change Pseudonym Result" },
    { UT_CHANGECURVATURE,               "Change Curvature" },
    { UT_CHANGESPEED,                   "Change Speed" },
    { UT_SETACCELERATIONCONTROLSTATUS,  "Set Acceleration Control Status" },
    { UT_SETEXTERIORLIGHTSSTATUS,       "Set Exterior Lights Status" },
    { UT_CHANGEHEADING,                 "Change Heading" },
    { UT_CHANGEYAWRATE,                 "Change YawRate" },
    { UT_SETDRIVEDIRECTION,             "Set Drive Direction" },
    { UT_SETEMBARKATIONSTATUS,          "Set Embarkation Status" },
    { UT_SETPTACTIVATION,               "Set Pt Activation" },
    { UT_CAMEVENTINDICATION,            "Cam Event Indication" },
    { UT_SETSTATIONTYPE,                "Set Station Type" },
    { UT_SETVEHICLEROLE,                "Set Vehicule Role" },
    { UT_SETDANGEROUSGOODS,             "Set Dangerous Goods" },
    { UT_SETDANGEROUSGOODSEXT,          "Set Dangerous Goods Extended" },
    { UT_SETLIGHTBARSIREN,              "Set Light Bar Siren" },
    { UT_CAMTRIGGER_RESULT,             "Cam Trigger Result" },
    { UT_DENMTRIGGER,                   "Denm Trigger" },
    { UT_DENMTRIGGER_RESULT,            "Denm Trigger Result" },
    { UT_DENMEVENT,                     "Denm Event" },
    { UT_DENMEVENT_RESULT,              "Denm Event Result" },
    { UT_TERMINATE_DENMEVENT,           "Denm Terminate Event" },
    { UT_TERMINATE_DENMEVENT_RESULT,    "Denm Terminate Event Result" },
    { UT_DENMEVENTINDICATION,           "Denm Event Indication" },
    { UT_GENERATEGEOUNICAST,            "GeoUnicast Trigger" },
    { UT_GENERATEGEOBROADCAST,          "GeoBroadcast Trigger" },
    { UT_GENERATEGEOANYCAST,            "GeoAnycast Trigger" },
    { UT_GENERATEGEOSHB,                "GeoSHB Trigger" },
    { UT_GENERATEGEOTSB,                "GeoTSB Trigger" },
    { UT_GNEVENTINDICATION,             "Geonetworking Event Indication" },
    { UT_GNTRIGGER_RESULT,              "Geonetworking Trigger Result" },
    { UT_MAPSPAT_TRIGGER_EVENT,         "MAP/SPaT Event Trigger" },
    { UT_MAPSPAT_TRIGGER_EVENT_RESULT,  "MAP/SPaT Trigger Result" },
    { 0, NULL}
};

static const value_string ut_result_names[] = {
    { IDX_RESULT_OK, "Succeed" },
    { IDX_RESULT_KO, "Failed" },
    { 0, NULL}
};

static const value_string ut_securedmode_names[] = {
    { IDX_SECUREDMODE_DISABLE,  "Disabled" },
    { IDX_SECUREDMODE_ENABLE,   "Enabled" },
    { 0, NULL}
};

static const value_string ut_drive_direction_names[] = {
  { 0x00, "Forward" },
  { 0x01, "Backward" },
  { 0x02, "Unavailable" },
  { 0, NULL}
};

static const value_string ut_station_type_names[] = {
  { 0, "Unknown" },
  { 1, "Pedestrian" },
  { 2, "Cyclist" },
  { 3, "Moped" },
  { 4, "Motorcycle" },
  { 5, "Passenger Car" },
  { 6, "Bus" },
  { 7, "Light Truck" },
  { 8, "Heavy Truck" },
  { 9, "Trailer" },
  { 10, "Special Vehicle" },
  { 11, "Tram" },
  { 12, "Reserved" },
  { 13, "Reserved" },
  { 14, "Reserved" },
  { 15, "Road Side Unit" },
  { 0, NULL}
};

static const value_string ut_vehicle_role_names[] = {
  { 0, "default" },
  { 1, "publicTransport" },
  { 2, "specialTransport" },
  { 3, "dangerousGoods" },
  { 4, "roadWork" },
  { 5, "rescue" },
  { 6, "emergency" },
  { 7, "safetyCar" },
  { 0, NULL}
};

static const value_string ut_embarkation_status_names[] = {
    { 0x00, "False" },
    { 0xff, "True" },
    { 0, NULL}
};

static const value_string ut_relevancetrafficdirection_names[] = {
    { 0x00, "all traffic directions" },
    { 0x01, "upstream traffic" },
    { 0x02, "downstream traffic" },
    { 0x03, "opposite traffic" },
    { 0, NULL}
};

static const value_string ut_relevancedistance_names[] = {
    { 0x00, "less than 50 m" },
    { 0x01, "less than 100 m" },
    { 0x02, "less than 200 m" },
    { 0x03, "less than 500 m" },
    { 0x04, "less than 1 000 m" },
    { 0x05, "less than 5 km" },
    { 0x06, "less than 10 km" },
    { 0x07, "greater than 10 km" },
    { 0, NULL}
};

static const value_string ut_setdangerousgoods_names[] = {
    { 0, "explosives1" },
    { 1, "explosives2" },
    { 2, "explosives3" },
    { 3, "explosives4" },
    { 4, "explosives5" },
    { 5, "explosives6" },
    { 6, "flammableGases" },
    { 7, "nonFlammableGases" },
    { 8, "toxicGases" },
    { 9, "flammableLiquids" },
    { 10, "flammableSolids" },
    { 11, "substancesLiableToSpontaneousCombustion" },
    { 12, "substancesEmittingFlammableGasesUponContactWithWater" },
    { 13, "oxidizingSubstances" },
    { 14, "organicPeroxides" },
    { 15, "toxicSubstances" },
    { 16, "infectiousSubstances" },
    { 17, "radioactiveMaterial" },
    { 18, "gcorrosiveSubstances" },
    { 19, "miscellaneousDangerousSubstances" },
    { 0, NULL}
};

static const value_string ut_shape_names[] = {
    { 0x00, "Circle" },
    { 0x01, "Rectangle" },
    { 0x02, "Ellipse" },
    { 0, NULL}
};

static const value_string ut_mapspat_event_names[] = {
    { 0x00, "Generate a SPAT message" },
    { 0x01, "Generate a MAP message with new content" },
    { 0x02, "Force MAP repetition < 10s" },
    { 0x03, "Force MAP repetition > 10s" },
    { 0, NULL}
};

/*****************************************************************************
    Command messages
*****************************************************************************/
/* Code to build tree for UtInitialise command */
static int dissect_ut_initialise(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
  guint64 hashedId8;

  /* Extract HashedId8 value */
  hashedId8 = tvb_get_ntoh64(tvb, offset + 1);
  
  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_initialize, tvb, offset, L_UT_INITIALISE, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtInitialize */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* UtInitialise tree */
  proto_tree_add_item(tree, hf_hashed_id8, tvb, offset, L_HASHEDID8, FALSE);
  offset += L_HASHEDID8; 

  return offset;
} // End of function dissect_ut_initialise

/* Code to build tree for UtChangePosition command */
static int dissect_ut_change_position(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_change_position, tvb, offset, L_UT_CHANGEPOSITION, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtInitialize */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* UtChangePosition tree */
  /* DeltaLatitude */
  proto_tree_add_item(tree, hf_delta_latitude, tvb, offset, L_DELTALATITUDE, FALSE);
  offset += L_DELTALATITUDE;
  /* DeltaLongitude */
  proto_tree_add_item(tree, hf_delta_longitude, tvb, offset, L_DELTALONGITUDE, FALSE);
  offset += L_DELTALONGITUDE;
  /* DeltaElevation */
  proto_tree_add_item(tree, hf_delta_elevation, tvb, offset, L_DELTAELEVATION, FALSE);
  offset += L_DELTAELEVATION;

  return offset;
} // End of function dissect_ut_change_position

/* Code to build tree for UtChangePseudonym command */
static int dissect_ut_change_pseudonym(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_change_pseudonym, tvb, offset, L_UT_CHANGEPSEUDONYM, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtInitialize */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  return offset;
} // End of function dissect_ut_change_pseudonym

/* Code to build tree for UtChangeCurvature command */
static int dissect_ut_change_curvature(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_change_curvature, tvb, offset, L_UT_CHANGECURVATURE, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtChangeSpeed */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* UtChangeSpeed tree */
  /* SpeedVariation */
  proto_tree_add_item(tree, hf_speed_variation, tvb, offset, L_CURVATURE, FALSE);
  offset += L_CURVATURE;

  return offset;
} // End of function dissect_ut_change_curvature

/* Code to build tree for UtChangeSpeed command */
static int dissect_ut_change_speed(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_change_speed, tvb, offset, L_UT_CHANGESPEED, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtChangeSpeed */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* UtChangeSpeed tree */
  /* SpeedVariation */
  proto_tree_add_item(tree, hf_speed_variation, tvb, offset, L_SPEEDVARIATION, FALSE);
  offset += L_SPEEDVARIATION;

  return offset;
} // End of function dissect_ut_change_speed

/* Code to build tree for UtSetAccelerationControlStatus command */
static int dissect_ut_set_acceleration_control_status(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_tree *tree_flags = NULL;
  proto_item *ti = NULL;
  proto_item *ti_flags = NULL;
  guint8 ut_flags = -1;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_set_acceleration_control_status, tvb, offset, L_UT_SETACCELERATIONCONTROLSTATUS, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtChangeSpeed */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;
  
  /* UtSetAccelerationControlStatus tree */
  /* flags */
  ti_flags = proto_tree_add_item(tree, hf_set_acceleration_control_status_flags, tvb, offset, L_UT_FLAGS, FALSE);
  tree_flags = proto_item_add_subtree(ti_flags, ett_ut_flags);
  ut_flags = tvb_get_guint8(tvb, offset);
  if ((ut_flags & 0x80) == 0x80) {
    proto_tree_add_item(tree_flags, hf_set_acceleration_control_status_flags_B_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x40) == 0x40) {
    proto_tree_add_item(tree_flags, hf_set_acceleration_control_status_flags_G_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x20) == 0x20) {
    proto_tree_add_item(tree_flags, hf_set_acceleration_control_status_flags_E_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x10) == 0x10) {
    proto_tree_add_item(tree_flags, hf_set_acceleration_control_status_flags_C_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x08) == 0x08) {
    proto_tree_add_item(tree_flags, hf_set_acceleration_control_status_flags_A_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x04) == 0x04) {
    proto_tree_add_item(tree_flags, hf_set_acceleration_control_status_flags_CC_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x02) == 0x02) {
    proto_tree_add_item(tree_flags, hf_set_acceleration_control_status_flags_L_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  offset += L_UT_FLAGS;
  
  return offset;
} // End of function dissect_ut_set_acceleration_control_status

/* Code to build tree for UtSetExteriorLightsStatus command */
static int dissect_ut_set_exterior_lights_status(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_tree *tree_flags = NULL;
  proto_item *ti = NULL;
  proto_item *ti_flags = NULL;
  guint8 ut_flags = -1;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_set_exterior_lights_status, tvb, offset, L_UT_SETEXTERIORLIGHTSSTATUS, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtChangeSpeed */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;
  
  /* UtSetAccelerationControlStatus tree */
  /* flags */
  ti_flags = proto_tree_add_item(tree, hf_set_exterior_lights_status_flags, tvb, offset, L_UT_FLAGS, FALSE);
  tree_flags = proto_item_add_subtree(ti_flags, ett_ut_flags);
  ut_flags = tvb_get_guint8(tvb, offset);
  if ((ut_flags & 0x80) == 0x80) {
    proto_tree_add_item(tree_flags, hf_set_exterior_lights_status_flags_LB_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x40) == 0x40) {
    proto_tree_add_item(tree_flags, hf_set_exterior_lights_status_flags_HB_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x20) == 0x20) {
    proto_tree_add_item(tree_flags, hf_set_exterior_lights_status_flags_LT_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x10) == 0x10) {
    proto_tree_add_item(tree_flags, hf_set_exterior_lights_status_flags_RT_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x08) == 0x08) {
    proto_tree_add_item(tree_flags, hf_set_exterior_lights_status_flags_D_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x04) == 0x04) {
    proto_tree_add_item(tree_flags, hf_set_exterior_lights_status_flags_R_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x02) == 0x02) {
    proto_tree_add_item(tree_flags, hf_set_exterior_lights_status_flags_F_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x01) == 0x01) {
    proto_tree_add_item(tree_flags, hf_set_exterior_lights_status_flags_P_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  offset += L_UT_FLAGS;
  
  return offset;
} // End of function dissect_ut_set_exterior_lights_status

/* Code to build tree for UtChangeHeading command */
static int dissect_ut_change_heading(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_change_heading, tvb, offset, L_UT_CHANGEHEADING, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtChangeHeading */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* UtChangeHeading tree */
  /* Direction */
  proto_tree_add_item(tree, hf_heading_direction, tvb, offset, L_HEADING_DIRECTION, FALSE);
  offset += L_HEADING_DIRECTION;

  return offset;
} // End of function dissect_ut_change_heading

/* Code to build tree for SetDriveDirection command */
static int dissect_ut_set_drive_direction(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_set_drive_direction, tvb, offset, L_UT_SETDRIVEDIRECTION, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* SetDriveDirection */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* SetDriveDirection tree */
  /* Direction */
  proto_tree_add_item(tree, hf_drive_direction, tvb, offset, L_DRIVE_DIRECTION, FALSE);
  offset += L_DRIVE_DIRECTION;

  return offset;
} // End of function dissect_ut_set_drive_direction

/* Code to build tree for ChangeYawRate command */
static int dissect_ut_change_yawrate(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_change_yawrate, tvb, offset, L_UT_CHANGEYAWRATE, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtChangeSpeed */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* ChangeYawRate tree */
  /* YawRate */
  proto_tree_add_item(tree, hf_yawrate, tvb, offset, L_YAWRATE, FALSE);
  offset += L_YAWRATE;

  return offset;
} // End of function dissect_ut_change_yawrate

/* Code to build tree for CamEventIndication command */
static int dissect_ut_cam_event_indication(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
  guint16 length = -1;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_cam_event_indication, tvb, offset, -1, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* CamEventIndication */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* CamEventIndication tree */
  /* Length */
  length = (guint16)((tvb_get_guint8(tvb, offset) << 8) & 0xff00) | (guint16)(tvb_get_guint8(tvb, offset + 1) & 0x00ff);
  proto_tree_add_item(tree, hf_cam_event_indication_length, tvb, offset, L_CAMEVENTLENGTH, FALSE);
  offset += L_CAMEVENTLENGTH;
  if (length != 0) {
    proto_tree_add_item(tree, hf_cam_event_indication_payload, tvb, offset, length, FALSE);
    offset += length;
  }

  return offset;
} // End of function dissect_ut_cam_event_indication

/* Code to build tree for SetStationType command */
static int dissect_ut_set_station_type(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_set_station_type, tvb, offset, L_UT_SETSTATIONTYPE, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* SetStationType */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* SetStationType tree */
  /* StationType */
  proto_tree_add_item(tree, hf_station_type, tvb, offset, L_STATIONTYPE, FALSE);
  offset += L_STATIONTYPE;

  return offset;
} // End of function dissect_ut_set_station_type

/* Code to build tree for SetVehicleRole command */
static int dissect_ut_set_vehicle_role(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_set_vehicle_role, tvb, offset, L_UT_SETVEHICLEROLE, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* SetVehicleRole */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* SetVehicleRole tree */
  /* VehicleRole */
  proto_tree_add_item(tree, hf_vehicle_role, tvb, offset, L_VEHICLEROLE, FALSE);
  offset += L_VEHICLEROLE;

  return offset;
} // End of function dissect_ut_set_vehicle_role

/* Code to build tree for SetEmbarkationStatus command */
static int dissect_ut_set_embarkation_status(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_set_embarkation_status, tvb, offset, L_UT_SETEMBARKATIONSTATUS, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* SetEmbarkationStatus */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* SetEmbarkationStatus tree */
  /* EmbarkationStatus */
  proto_tree_add_item(tree, hf_embarkation_status, tvb, offset, L_EMBARKATIONSTATUS, FALSE);
  offset += L_EMBARKATIONSTATUS;

  return offset;
} // End of function dissect_ut_set_embarkation_status

/* Code to build tree for SetPtActivation command */
static int dissect_ut_set_pt_activation(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
  guint16 length = -1;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_set_pt_activation, tvb, offset, -1, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* SetPtActivation */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* SetPtActivation tree */
  /* Length */
  length = tvb_get_guint8(tvb, offset);
  proto_tree_add_item(tree, hf_set_pt_activation_length, tvb, offset, L_SETPTACTIVATIONLENGTH, FALSE);
  offset += L_SETPTACTIVATIONLENGTH;
  if (length != 0) {
    proto_tree_add_item(tree, hf_set_pt_activation_payload, tvb, offset, length, FALSE);
    offset += length;
  }

  return offset;
} // End of function dissect_ut_set_pt_activation

/* Code to build tree for SetDangerousGoods command */
static int dissect_ut_set_dangerous_goods(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_set_dangerous_goods, tvb, offset, L_UT_SETDANGEROUSGOODS, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* SetDangerousGoods */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* SetDangerousGoods tree */
  /* DangerousGoods */
  proto_tree_add_item(tree, hf_dangerous_goods, tvb, offset, L_DANGEROUSGOODS, FALSE);
  offset += L_DANGEROUSGOODS;

  return offset;
} // End of function dissect_ut_set_dangerous_goods

/* Code to build tree for SetDangerousGoodsExtended command */
static int dissect_ut_set_dangerous_goods_ext(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_set_dangerous_goods_ext, tvb, offset, L_UT_SETDANGEROUSGOODSEXT, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* SetDangerousGoodsExtended */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* SetDangerousGoodsExtended tree */
  /* DangerousGoodsExtended */
  proto_tree_add_item(tree, hf_dangerous_goods_ext, tvb, offset, L_DANGEROUSGOODS, FALSE);
  offset += L_DANGEROUSGOODS;

  return offset;
} // End of function dissect_ut_set_dangerous_goods

/* Code to build tree for UtSetLightBarSiren command */
static int dissect_ut_set_light_bar_siren(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_tree *tree_flags = NULL;
  proto_item *ti = NULL;
  proto_item *ti_flags = NULL;
  guint8 ut_flags = -1;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_set_light_bar_siren, tvb, offset, L_UT_SETLIGHTBARSIREN, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtChangeSpeed */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;
  
  /* SetLightBarSiren tree */
  /* flags */
  ti_flags = proto_tree_add_item(tree, hf_set_light_bar_siren_flags, tvb, offset, L_UT_FLAGS, FALSE);
  tree_flags = proto_item_add_subtree(ti_flags, ett_ut_flags);
  ut_flags = tvb_get_guint8(tvb, offset);
  if ((ut_flags & 0x80) == 0x80) {
    proto_tree_add_item(tree_flags, hf_set_light_bar_siren_flags_LB_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x40) == 0x40) {
    proto_tree_add_item(tree_flags, hf_set_light_bar_siren_flags_S_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  offset += L_UT_FLAGS;
  
  return offset;
} // End of function dissect_ut_set_light_bar_siren

/* Code to build tree for UtDenmTrigger command */
static int dissect_ut_denm_trigger(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
  proto_tree *tree_flags = NULL;
  proto_item *ti_flags = NULL;
  guint8 ut_flags = -1;
  guint8 alacarteLength = -1;
   
  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_denm_trigger, tvb, offset, -1, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtDenmTrigger */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* UtDenmTrigger tree */
  /* flags */
  ti_flags = proto_tree_add_item(tree, hf_denm_trigger_flags, tvb, offset, L_UT_FLAGS, FALSE);
  tree_flags = proto_item_add_subtree(ti_flags, ett_ut_flags);
  ut_flags = tvb_get_guint8(tvb, offset);
  if ((ut_flags & 0x80) == 0x80) {
    proto_tree_add_item(tree_flags, hf_denm_trigger_flags_V_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x40) == 0x40) {
    proto_tree_add_item(tree_flags, hf_denm_trigger_flags_R_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x04) == 0x04) {
    proto_tree_add_item(tree_flags, hf_denm_trigger_flags_K_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  offset += L_UT_FLAGS;
  /* DetectionTime */
  proto_tree_add_item(tree, hf_detection_time, tvb, offset, L_DETECTIONTIME, FALSE);
  offset += L_DETECTIONTIME;
  /* ValidityDuration */
  proto_tree_add_item(tree, hf_validity_duration, tvb, offset, L_VALIDITYDURATION, FALSE);
  offset += L_VALIDITYDURATION;
  /* RepetitionDuration */
  proto_tree_add_item(tree, hf_repetition_duration, tvb, offset, L_REPETITIONDURATION, FALSE);
  offset += L_REPETITIONDURATION;
  /* InfoQuality */
  proto_tree_add_item(tree, hf_info_quality, tvb, offset, L_INFOQUALITY, FALSE);
  offset += L_INFOQUALITY;
  /* Cause */
  proto_tree_add_item(tree, hf_cause, tvb, offset, L_CAUSE, FALSE);
  offset += L_CAUSE;
  /* SubCause */
  proto_tree_add_item(tree, hf_subcause, tvb, offset, L_SUBCAUSE, FALSE);
  offset += L_SUBCAUSE;
  /* RelevanceDistance */
  proto_tree_add_item(tree, hf_relevance_distance, tvb, offset, L_RELEVANCEDISTANCE, FALSE);
  offset += L_RELEVANCEDISTANCE;
  /* RelevanceTrafficDirection */
  proto_tree_add_item(tree, hf_relevance_traffic_direction, tvb, offset, L_RELEVANCETRAFFICDIRECTION, FALSE);
  offset += L_RELEVANCETRAFFICDIRECTION;
  /* TransmissionInterval */
  proto_tree_add_item(tree, hf_transmission_interval, tvb, offset, L_TRANSMISSIONINTERVAL, FALSE);
  offset += L_TRANSMISSIONINTERVAL;
  /* RepetitionInterval */
  proto_tree_add_item(tree, hf_repetition_interval, tvb, offset, L_REPETITIONINTERVAL, FALSE);
  offset += L_REPETITIONINTERVAL;
  /* alacarteLength */
  alacarteLength = tvb_get_guint8(tvb, offset);
  proto_tree_add_item(tree, hf_alacarte_length, tvb, offset, L_ALACARTELENGTH, FALSE);
  offset += L_ALACARTELENGTH;
  if (alacarteLength != 0) {
    proto_tree_add_item(tree, hf_alacarte, tvb, offset, alacarteLength, FALSE);
    offset += alacarteLength;
  }
  
  return offset;
} // End of function dissect_ut_denm_trigger

/* Code to build tree for UtDenmEvent command */
static int dissect_ut_denm_event(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
  proto_tree *tree_flags = NULL;
  proto_item *ti_flags = NULL;
  guint8 ut_flags = -1;
  guint8 alacarteLength = -1;
   
  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_denm_event, tvb, offset, -1, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtDenmTrigger */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* UtDenmEvent tree */
  /* flags */
  ti_flags = proto_tree_add_item(tree, hf_denm_event_flags, tvb, offset, L_UT_FLAGS, FALSE);
  tree_flags = proto_item_add_subtree(ti_flags, ett_ut_flags);
  ut_flags = tvb_get_guint8(tvb, offset);
  if ((ut_flags & 0x80) == 0x80) {
    proto_tree_add_item(tree_flags, hf_denm_event_flags_V_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x40) == 0x40) {
    proto_tree_add_item(tree_flags, hf_denm_event_flags_S_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x20) == 0x20) {
    proto_tree_add_item(tree_flags, hf_denm_event_flags_D_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x10) == 0x10) {
    proto_tree_add_item(tree_flags, hf_denm_event_flags_T_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x08) == 0x08) {
    proto_tree_add_item(tree_flags, hf_denm_event_flags_C_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x04) == 0x04) {
    proto_tree_add_item(tree_flags, hf_denm_event_flags_K_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x02) == 0x02) {
    proto_tree_add_item(tree_flags, hf_denm_event_flags_I_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  if ((ut_flags & 0x01) == 0x01) {
    proto_tree_add_item(tree_flags, hf_denm_event_flags_X_bit, tvb, offset, L_UT_FLAGS, FALSE);
  }
  offset += L_UT_FLAGS;
  /* StationID */
  proto_tree_add_item(tree, hf_station_id, tvb, offset, L_STATIONID, FALSE);
  offset += L_STATIONID;
  /* SequenceNumber */
  proto_tree_add_item(tree, hf_sequence_number, tvb, offset, L_SEQNUM, FALSE);
  offset += L_SEQNUM;
  /* DetectionTime */
  proto_tree_add_item(tree, hf_detection_time, tvb, offset, L_DETECTIONTIME, FALSE);
  offset += L_DETECTIONTIME;
  /* ValidityDuration */
  proto_tree_add_item(tree, hf_validity_duration, tvb, offset, L_VALIDITYDURATION, FALSE);
  offset += L_VALIDITYDURATION;
  /* InfoQuality */
  proto_tree_add_item(tree, hf_info_quality, tvb, offset, L_INFOQUALITY, FALSE);
  offset += L_INFOQUALITY;
  /* Cause */
  proto_tree_add_item(tree, hf_cause, tvb, offset, L_CAUSE, FALSE);
  offset += L_CAUSE;
  /* SubCause */
  proto_tree_add_item(tree, hf_subcause, tvb, offset, L_SUBCAUSE, FALSE);
  offset += L_SUBCAUSE;
  /* RelevanceDistance */
  proto_tree_add_item(tree, hf_relevance_distance, tvb, offset, L_RELEVANCEDISTANCE, FALSE);
  offset += L_RELEVANCEDISTANCE;
  /* RelevanceTrafficDirection */
  proto_tree_add_item(tree, hf_relevance_traffic_direction, tvb, offset, L_RELEVANCETRAFFICDIRECTION, FALSE);
  offset += L_RELEVANCETRAFFICDIRECTION;
  /* TransmissionInterval */
  proto_tree_add_item(tree, hf_transmission_interval, tvb, offset, L_TRANSMISSIONINTERVAL, FALSE);
  offset += L_TRANSMISSIONINTERVAL;
  /* RepetitionInterval */
  proto_tree_add_item(tree, hf_repetition_interval, tvb, offset, L_REPETITIONINTERVAL, FALSE);
  offset += L_REPETITIONINTERVAL;
  /* alacarteLength */
  alacarteLength = tvb_get_guint8(tvb, offset);
  proto_tree_add_item(tree, hf_alacarte_length, tvb, offset, L_ALACARTELENGTH, FALSE);
  offset += L_ALACARTELENGTH;
  if (alacarteLength != 0) {
    proto_tree_add_item(tree, hf_alacarte, tvb, offset, alacarteLength, FALSE);
    offset += alacarteLength;
  }
  
  return offset;
} // End of function dissect_ut_denm_event

/* Code to build tree for UtTerminateDenmEvent indication */
static int dissect_ut_terminate_denm_event(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_terminate_denm_event, tvb, offset, L_UT_INITIALISE_RESULT, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtTerminateDenmEvent */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* StationID */
  proto_tree_add_item(tree, hf_station_id, tvb, offset, L_STATIONID, FALSE);
  offset += L_STATIONID;

  /* SequenceNumber */
  proto_tree_add_item(tree, hf_sequence_number, tvb, offset, L_SEQNUM, FALSE);
  offset += L_SEQNUM;

  return offset;
} // End of function dissect_ut_terminate_denm_event

/* Code to build tree for DenmEventIndication command */
static int dissect_ut_denm_event_indication(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
  guint16 length = -1;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_denm_event_indication, tvb, offset, -1, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* DenmEventIndication */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* DenmEventIndication tree */
  /* Length */
  /* length */
  length = (guint16)((tvb_get_guint8(tvb, offset) << 8) & 0xff00) | (guint16)(tvb_get_guint8(tvb, offset + 1) & 0x00ff);
  proto_tree_add_item(tree, hf_denm_event_indication_length, tvb, offset, L_DENMEVENTLENGTH, FALSE);
  offset += L_DENMEVENTLENGTH;
  if (length != 0) {
    proto_tree_add_item(tree, hf_denm_event_indication_payload, tvb, offset, length, FALSE);
    offset += length;
  }

  return offset;
} // End of function dissect_ut_denm_event_indication

/* Code to build tree for GnGenerateGeoUnicast command */
static int dissect_ut_gn_geounicast(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
  guint16 payload_length = -1;
   
  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_gn_geounicast, tvb, offset, -1, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* GnGenerateGeoUnicast */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* GnGenerateGeoUnicast tree */
  /* DstGnAddr */
  proto_tree_add_item(tree, hf_gn_address, tvb, offset, L_DSTGNADDR, FALSE);
  offset += L_DSTGNADDR;
  /* Lifetime */
  proto_tree_add_item(tree, hf_lifetime, tvb, offset, L_LIFETIME, FALSE);
  offset += L_LIFETIME;
  /* TrafficClass */
  proto_tree_add_item(tree, hf_trafficclass, tvb, offset, L_TRAFFICCLASS, FALSE);
  offset += L_TRAFFICCLASS;
  /* PayloadLength */
  if (offset < (int)tvb_length(tvb)) {
  payload_length = (guint16)((tvb_get_guint8(tvb, offset) << 8) & 0xff00) | (guint16)(tvb_get_guint8(tvb, offset + 1) & 0x00ff);
  proto_tree_add_item(tree, hf_payloadlength, tvb, offset, L_PAYLOADLENGTH, FALSE);
  offset += L_PAYLOADLENGTH;
  if (payload_length != 0) {
    proto_tree_add_item(tree, hf_payload, tvb, offset, payload_length, FALSE);
    offset += payload_length;
  }
  }
  
  return offset;
} // End of function dissect_ut_gn_geounicast

/* Code to build tree for GnGenerateGeoBroadcast command */
static int dissect_ut_gn_geobroadcast(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
  guint16 payload_length = -1;
   
  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_gn_geobroadcast, tvb, offset, -1, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* GnGenerateGeoBroadcast */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* GnGenerateGeoBroadcast tree */
  /* Shape */
  proto_tree_add_item(tree, hf_station_id, tvb, offset, L_SHAPE, FALSE);
  offset += L_SHAPE;
  /* Lifetime */
  proto_tree_add_item(tree, hf_lifetime, tvb, offset, L_LIFETIME, FALSE);
  offset += L_LIFETIME;
  /* TrafficClass */
  proto_tree_add_item(tree, hf_trafficclass, tvb, offset, L_TRAFFICCLASS, FALSE);
  offset += L_TRAFFICCLASS;
  /* Reserved */
  proto_tree_add_item(tree, hf_reservedgeobroadcast, tvb, offset, L_RESERVEDGEOBROADCAST, FALSE);
  offset += L_RESERVEDGEOBROADCAST;
  /* Latitude */
  proto_tree_add_item(tree, hf_latitude, tvb, offset, L_LATITUDE, FALSE);
  offset += L_LATITUDE;
  /* Longitude */
  proto_tree_add_item(tree, hf_longitude, tvb, offset, L_LONGITUDE, FALSE);
  offset += L_LONGITUDE;
  /* DistanceA */
  proto_tree_add_item(tree, hf_distancea, tvb, offset, L_DISTANCEA, FALSE);
  offset += L_DISTANCEA;
  /* DistanceB */
  proto_tree_add_item(tree, hf_distanceb, tvb, offset, L_DISTANCEB, FALSE);
  offset += L_DISTANCEB;
  /* Angle */
  proto_tree_add_item(tree, hf_angle, tvb, offset, L_ANGLE, FALSE);
  offset += L_ANGLE;
  /* PayloadLength */
  if (offset < (int)tvb_length(tvb)) {
  payload_length = (guint16)((tvb_get_guint8(tvb, offset) << 8) & 0xff00) | (guint16)(tvb_get_guint8(tvb, offset + 1) & 0x00ff);
  proto_tree_add_item(tree, hf_payloadlength, tvb, offset, L_PAYLOADLENGTH, FALSE);
  offset += L_PAYLOADLENGTH;
  if (payload_length != 0) {
    proto_tree_add_item(tree, hf_payload, tvb, offset, payload_length, FALSE);
    offset += payload_length;
  }
  }
  return offset;
} // End of function dissect_ut_gn_geobroadcast

/* Code to build tree for GnGenerateGeoAnycast command */
static int dissect_ut_gn_geoanycast(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
  guint16 payload_length = -1;
   
  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_gn_geoanycast, tvb, offset, -1, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* GnGenerateGeoAnycast */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* GnGenerateGeoAnycast tree */
  /* Shape */
  proto_tree_add_item(tree, hf_station_id, tvb, offset, L_SHAPE, FALSE);
  offset += L_SHAPE;
  /* Lifetime */
  proto_tree_add_item(tree, hf_lifetime, tvb, offset, L_LIFETIME, FALSE);
  offset += L_LIFETIME;
  /* TrafficClass */
  proto_tree_add_item(tree, hf_trafficclass, tvb, offset, L_TRAFFICCLASS, FALSE);
  offset += L_TRAFFICCLASS;
  /* Reserved */
  proto_tree_add_item(tree, hf_reservedgeobroadcast, tvb, offset, L_RESERVEDGEOBROADCAST, FALSE);
  offset += L_RESERVEDGEOBROADCAST;
  /* Latitude */
  proto_tree_add_item(tree, hf_latitude, tvb, offset, L_LATITUDE, FALSE);
  offset += L_LATITUDE;
  /* Longitude */
  proto_tree_add_item(tree, hf_longitude, tvb, offset, L_LONGITUDE, FALSE);
  offset += L_LONGITUDE;
  /* DistanceA */
  proto_tree_add_item(tree, hf_distancea, tvb, offset, L_DISTANCEA, FALSE);
  offset += L_DISTANCEA;
  /* DistanceB */
  proto_tree_add_item(tree, hf_distanceb, tvb, offset, L_DISTANCEB, FALSE);
  offset += L_DISTANCEB;
  /* Angle */
  proto_tree_add_item(tree, hf_angle, tvb, offset, L_ANGLE, FALSE);
  offset += L_ANGLE;
  /* PayloadLength */
  if (offset < (int)tvb_length(tvb)) {
  payload_length = (guint16)((tvb_get_guint8(tvb, offset) << 8) & 0xff00) | (guint16)(tvb_get_guint8(tvb, offset + 1) & 0x00ff);
  proto_tree_add_item(tree, hf_payloadlength, tvb, offset, L_PAYLOADLENGTH, FALSE);
  offset += L_PAYLOADLENGTH;
  if (payload_length != 0) {
    proto_tree_add_item(tree, hf_payload, tvb, offset, payload_length, FALSE);
    offset += payload_length;
  }
  }
  
  return offset;
} // End of function dissect_ut_gn_geobroadcast

/* Code to build tree for GnGenerateGeoShb command */
static int dissect_ut_gn_geoshb(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
  guint16 payload_length = -1;
   
  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_gn_geoshb, tvb, offset, -1, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* GnGenerateGeoShb */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* GnGenerateGeoShb tree */
  /* TrafficClass */
  proto_tree_add_item(tree, hf_trafficclass, tvb, offset, L_TRAFFICCLASS, FALSE);
  offset += L_TRAFFICCLASS;
  /* PayloadLength */
  payload_length = (guint16)((tvb_get_guint8(tvb, offset) << 8) & 0xff00) | (guint16)(tvb_get_guint8(tvb, offset + 1) & 0x00ff);
  proto_tree_add_item(tree, hf_payloadlength, tvb, offset, L_PAYLOADLENGTH, FALSE);
  offset += L_PAYLOADLENGTH;
  if (payload_length != 0) {
    proto_tree_add_item(tree, hf_payload, tvb, offset, payload_length, FALSE);
    offset += payload_length;
  }
  
  return offset;
} // End of function dissect_ut_gn_geoshb

/* Code to build tree for GnGenerateGeoTsb command */
static int dissect_ut_gn_geotsb(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
  guint16 payload_length = -1;
   
  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_gn_geotsb, tvb, offset, -1, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* GnGenerateGeoTsb */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* GnGenerateGeoTsb tree */
  /* NbHops */
  proto_tree_add_item(tree, hf_trafficclass, tvb, offset, L_NBHOPS, FALSE);
  offset += L_NBHOPS;
  /* TrafficClass */
  proto_tree_add_item(tree, hf_trafficclass, tvb, offset, L_TRAFFICCLASS, FALSE);
  offset += L_TRAFFICCLASS;
  /* PayloadLength */
  if (offset < (int)tvb_length(tvb)) {
  payload_length = (guint16)((tvb_get_guint8(tvb, offset) << 8) & 0xff00) | (guint16)(tvb_get_guint8(tvb, offset + 1) & 0x00ff);
  proto_tree_add_item(tree, hf_payloadlength, tvb, offset, L_PAYLOADLENGTH, FALSE);
  offset += L_PAYLOADLENGTH;
  if (payload_length != 0) {
    proto_tree_add_item(tree, hf_payload, tvb, offset, payload_length, FALSE);
    offset += payload_length;
  }
  }
  
  return offset;
} // End of function dissect_ut_gn_geotsb

/* Code to build tree for GnEventIndication command */
static int dissect_ut_gn_event_indication(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
  guint16 length = -1;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_gn_event_indication, tvb, offset, -1, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* GnEventIndication */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* GnEventIndication tree */
  /* Length */
  length = (guint16)((tvb_get_guint8(tvb, offset) << 8) & 0xff00) | (guint16)(tvb_get_guint8(tvb, offset + 1) & 0x00ff);
  proto_tree_add_item(tree, hf_gn_event_indication_length, tvb, offset, L_PAYLOADLENGTH, FALSE);
  offset += L_PAYLOADLENGTH;
  if (length != 0) {
    proto_tree_add_item(tree, hf_gn_event_indication_payload, tvb, offset, length, FALSE);
    offset += length;
  }

  return offset;
} // End of function dissect_ut_gn_event_indication

/* Code to build tree for UtMapSpatTrigger command */
static int dissect_ut_mapspat_trigger_event(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
   
  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_mapspat_trigger_event, tvb, offset, L_UT_MAPSPAT_TRIGGER_EVENT, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtMapSpatTrigger */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;

  /* UtMapSpatTrigger tree */
  /* Event */
  proto_tree_add_item(tree, hf_event, tvb, offset, L_MAPSPAT_TRIGGER_EVENT, FALSE);
  offset += L_MAPSPAT_TRIGGER_EVENT;
  
  return offset;
} // End of function dissect_ut_mapspat_trigger_event

/*****************************************************************************
    Result messages
*****************************************************************************/
/* Code to build tree for UtInitialiseResult indication */
static int dissect_ut_initialise_result(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
  guint8 ut_code = -1;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_initialize_result, tvb, offset, L_UT_INITIALISE_RESULT, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtInitialiseResult */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;
  
  /* Result */
  ut_code = tvb_get_guint8(tvb, offset);
  proto_tree_add_item(tree, hf_initialize_result_code, tvb, offset, L_UT_INITIALISE_RESULT_CODE, FALSE);
  offset += L_UT_INITIALISE_RESULT_CODE;

  return offset;
} // End of function dissect_ut_initialise_result

/* Code to build tree for UtChangePositionResult indication */
static int dissect_ut_change_position_result(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
  guint8 ut_code = -1;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_change_position_result, tvb, offset, L_UT_CHANGEPOSITION_RESULT, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtInitialiseResult */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;
  
  /* Result */
  ut_code = tvb_get_guint8(tvb, offset);
  proto_tree_add_item(tree, hf_change_position_result_code, tvb, offset, L_UT_CHANGEPOSITION_RESULT_CODE, FALSE);
  offset += L_UT_CHANGEPOSITION_RESULT_CODE;

  return offset;
} // End of function dissect_ut_change_position_result

/* Code to build tree for UtChangePseudonymResult indication */
static int dissect_ut_change_pseudonym_result(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
  guint8 ut_code = -1;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_change_pseudonym_result, tvb, offset, L_UT_CHANGEPSEUDONYM_RESULT, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtInitialiseResult */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;
  
  /* Result */
  ut_code = tvb_get_guint8(tvb, offset);
  proto_tree_add_item(tree, hf_change_pseudonym_result_code, tvb, offset, L_UT_CHANGEPSEUDONYM_RESULT_CODE, FALSE);
  offset += L_UT_CHANGEPSEUDONYM_RESULT_CODE;

  return offset;
} // End of function dissect_ut_change_pseudonym_result

/* Code to build tree for UtCamTriggerResult indication */
static int dissect_ut_cam_trigger_result(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
  guint8 ut_code = -1;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_cam_trigger_result, tvb, offset, L_UT_CAM_TRIGGER_RESULT, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtCamTriggerResult */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;
  
  /* Result */
  ut_code = tvb_get_guint8(tvb, offset);
  proto_tree_add_item(tree, hf_cam_trigger_result_code, tvb, offset, L_UT_CAM_TRIGGER_RESULT_CODE, FALSE);
  offset += L_UT_CAM_TRIGGER_RESULT_CODE;

  return offset;
} // End of function dissect_ut_cam_trigger_result

/* Code to build tree for UtDenmTriggerResult indication */
static int dissect_ut_denm_trigger_result(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
  guint8 ut_code = -1;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_denm_trigger_result, tvb, offset, L_UT_DENMM_TRIGGER_RESULT, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtInitialiseResult */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;
  
  /* Result */
  ut_code = tvb_get_guint8(tvb, offset);
  proto_tree_add_item(tree, hf_denm_trigger_result_code, tvb, offset, L_UT_DENMM_TRIGGER_RESULT_CODE, FALSE);
  offset += L_UT_DENMM_TRIGGER_RESULT_CODE;

  /* StationID */
  proto_tree_add_item(tree, hf_station_id, tvb, offset, L_STATIONID, FALSE);
  offset += L_STATIONID;

  /* SequenceNumber */
  proto_tree_add_item(tree, hf_sequence_number, tvb, offset, L_SEQNUM, FALSE);
  offset += L_SEQNUM;

  return offset;
} // End of function dissect_ut_denm_trigger_result

/* Code to build tree for UtDenmEventResult indication */
static int dissect_ut_denm_event_result(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
  guint8 ut_code = -1;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_denm_event_result, tvb, offset, L_UT_DENMM_EVENT_RESULT, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtInitialiseResult */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;
  
  /* Result */
  ut_code = tvb_get_guint8(tvb, offset);
  proto_tree_add_item(tree, hf_denm_event_result_code, tvb, offset, L_UT_DENMM_EVENT_RESULT_CODE, FALSE);
  offset += L_UT_DENMM_EVENT_RESULT_CODE;

  /* StationID */
  proto_tree_add_item(tree, hf_station_id, tvb, offset, L_STATIONID, FALSE);
  offset += L_STATIONID;

  /* SequenceNumber */
  proto_tree_add_item(tree, hf_sequence_number, tvb, offset, L_SEQNUM, FALSE);
  offset += L_SEQNUM;

  return offset;
} // End of function dissect_ut_denm_event_result

/* Code to build tree for UtTerminateDenmEventResult indication */
static int dissect_ut_terminate_denm_event_result(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
  guint8 ut_code = -1;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_terminate_denm_event_result, tvb, offset, L_UT_INITIALISE_RESULT, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtTerminateDenmEventResult */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;
  
  /* Result */
  ut_code = tvb_get_guint8(tvb, offset);
  proto_tree_add_item(tree, hf_denm_terminate_event_result_code, tvb, offset, L_UT_INITIALISE_RESULT_CODE, FALSE);
  offset += L_UT_INITIALISE_RESULT_CODE;

  return offset;
} // End of function dissect_ut_terminate_denm_event_result

/* Code to build tree for UtGnTriggerResult indication */
static int dissect_ut_gn_trigger_result(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;
  guint8 ut_code = -1;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_gntrigger_result, tvb, offset, L_UT_GNTRIGGER_RESULT, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtGnTriggerResult */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;
  
  /* Result */
  ut_code = tvb_get_guint8(tvb, offset);
  proto_tree_add_item(tree, hf_gntrigger_result_code, tvb, offset, L_UT_INITIALISE_RESULT_CODE, FALSE);
  offset += L_UT_INITIALISE_RESULT_CODE;

  return offset;
} // End of function dissect_ut_gn_trigger_result

/* Code to build tree for UtMapSpatTriggerResult  indication */
static int dissect_ut_mapspat_trigger_event_result(tvbuff_t *tvb, proto_tree *header_tree, int offset)
{
  proto_tree *tree = NULL;
  proto_item *ti = NULL;

  /* UT command tree */
  ti = proto_tree_add_item(header_tree, hf_mapspat_trigger_event_result, tvb, offset, L_UT_MAPSPAT_TRIGGER_EVENT_RESULT, FALSE);
  tree = proto_item_add_subtree(ti, ett_ut_command);

  /* UtGnTriggerResult */
  proto_tree_add_item(tree, hf_command, tvb, offset, L_UT_COMMAND, FALSE);
  offset += L_UT_COMMAND;
  
  /* Result */
  proto_tree_add_item(tree, hf_mapspat_trigger_event_result_code, tvb, offset, L_UT_MAPSPAT_TRIGGER_EVENT_RESULT_CODE, FALSE);
  offset += L_UT_MAPSPAT_TRIGGER_EVENT_RESULT_CODE;

  /* Message count */
  proto_tree_add_item(tree, hf_mapspat_trigger_event_result_msgc, tvb, offset, L_UT_MAPSPAT_TRIGGER_EVENT_RESULT_MSGC, FALSE);
  offset += L_UT_MAPSPAT_TRIGGER_EVENT_RESULT_MSGC;

  return offset;
} // End of function dissect_ut_mapspat_trigger_event_result

/* Code to actually dissect the UT packets */
static int dissect_ut_packet(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  guint32 offset = 0;
  guint8 ut_command = -1;
  
  /* Check that there's enough data */
  if(tvb_length(tvb) < MIN_PDU_SIZE)
    return 0;

  /* Update COL_PROTOCOL */ 
  col_set_str(pinfo->cinfo, COL_PROTOCOL, "UpperTester Protocol");
  /* Clear out stuff in the info column */
  col_clear(pinfo->cinfo,COL_INFO);

  /* Extract the command */
  ut_command = tvb_get_guint8(tvb, offset);
    
  if (tree) { /* we are being asked for details */
    proto_item *ti = NULL;
    proto_tree *ut_tree = NULL;
    
    /* Main UT tree */
    ti = proto_tree_add_item(tree, proto_ut, tvb, offset, -1, FALSE);
    proto_item_append_text(ti, ": %s", val_to_str(ut_command, ut_command_names, "Unknown (0x%02x)"));
//    proto_item_append_text(ti, ": %s - Secured mode: %s", val_to_str(ut_command, ut_command_names, "Unknown (0x%02x)"), val_to_str(SecuredMode, ut_securedmode_names, "Unknown (0x%02x)"));
    ut_tree = proto_item_add_subtree(ti, ett_ut);

    /* Dissect */
    switch (ut_command) {
        case (guint8)UT_INITIALISE:
            offset = dissect_ut_initialise(tvb, ut_tree, offset);
            break;
        case (guint8)UT_INITIALISE_RESULT:
            offset = dissect_ut_initialise_result(tvb, ut_tree, offset);
            break;
        case (guint8)UT_CHANGEPOSITION:
            offset = dissect_ut_change_position(tvb, ut_tree, offset);
            break;
        case (guint8)UT_CHANGEPOSITION_RESULT:
            offset = dissect_ut_change_position_result(tvb, ut_tree, offset);
            break;
        case (guint8)UT_CHANGEPSEUDONYM:
            offset = dissect_ut_change_pseudonym(tvb, ut_tree, offset);
            break;
        case (guint8)UT_CHANGEPSEUDONYM_RESULT:
            offset = dissect_ut_change_pseudonym_result(tvb, ut_tree, offset);
            break;
        case (guint8)UT_CHANGECURVATURE:
            offset = dissect_ut_change_curvature(tvb, ut_tree, offset);
            break;
        case (guint8)UT_CHANGESPEED:
            offset = dissect_ut_change_speed(tvb, ut_tree, offset);
            break;
        case (guint8)UT_SETACCELERATIONCONTROLSTATUS:
            offset = dissect_ut_set_acceleration_control_status(tvb, ut_tree, offset);
            break;
        case (guint8)UT_SETEXTERIORLIGHTSSTATUS:
            offset = dissect_ut_set_exterior_lights_status(tvb, ut_tree, offset);
            break;
        case (guint8)UT_CHANGEHEADING:
            offset = dissect_ut_change_heading(tvb, ut_tree, offset);
            break;
        case (guint8)UT_SETDRIVEDIRECTION:
            offset = dissect_ut_set_drive_direction(tvb, ut_tree, offset);
            break;
        case (guint8)UT_CHANGEYAWRATE:
            offset = dissect_ut_change_yawrate(tvb, ut_tree, offset);
            break;
        case (guint8)UT_SETSTATIONTYPE:
            offset = dissect_ut_set_station_type(tvb, ut_tree, offset);
            break;
        case (guint8)UT_SETVEHICLEROLE:
            offset = dissect_ut_set_vehicle_role(tvb, ut_tree, offset);
            break;
        case (guint8)UT_SETEMBARKATIONSTATUS:
            offset = dissect_ut_set_embarkation_status(tvb, ut_tree, offset);
            break;
        case (guint8)UT_SETPTACTIVATION:
            offset = dissect_ut_set_pt_activation(tvb, ut_tree, offset);
            break;
        case (guint8)UT_CAMEVENTINDICATION:
            offset = dissect_ut_cam_event_indication(tvb, ut_tree, offset);
            break;
            offset = dissect_ut_set_dangerous_goods(tvb, ut_tree, offset);
            break;
        case (guint8)UT_SETDANGEROUSGOODSEXT:
            offset = dissect_ut_set_dangerous_goods_ext(tvb, ut_tree, offset);
            break;
        case (guint8)UT_SETLIGHTBARSIREN:
            offset = dissect_ut_set_light_bar_siren(tvb, ut_tree, offset);
            break;
        case (guint8)UT_CAMTRIGGER_RESULT:
            offset = dissect_ut_cam_trigger_result(tvb, ut_tree, offset);
            break;
        case (guint8)UT_DENMTRIGGER:
            offset = dissect_ut_denm_trigger(tvb, ut_tree, offset);
            break;
        case (guint8)UT_DENMTRIGGER_RESULT:
            offset = dissect_ut_denm_trigger_result(tvb, ut_tree, offset);
            break;
        case (guint8)UT_DENMEVENT:
            offset = dissect_ut_denm_event(tvb, ut_tree, offset);
            break;
        case (guint8)UT_DENMEVENT_RESULT:
            offset = dissect_ut_denm_event_result(tvb, ut_tree, offset);
            break;
        case (guint8)UT_TERMINATE_DENMEVENT:
            offset = dissect_ut_terminate_denm_event(tvb, ut_tree, offset);
            break;
        case (guint8)UT_DENMEVENTINDICATION:
            offset = dissect_ut_denm_event_indication(tvb, ut_tree, offset);
            break;
        case (guint8)UT_TERMINATE_DENMEVENT_RESULT:
            offset = dissect_ut_terminate_denm_event_result(tvb, ut_tree, offset);
            break;
        case (guint8)UT_GENERATEGEOUNICAST:
            offset = dissect_ut_gn_geounicast(tvb, ut_tree, offset);
            break;
        case (guint8)UT_GENERATEGEOBROADCAST:
            offset = dissect_ut_gn_geobroadcast(tvb, ut_tree, offset);
            break;
        case (guint8)UT_GENERATEGEOANYCAST:
            offset = dissect_ut_gn_geoanycast(tvb, ut_tree, offset);
            break;
        case (guint8)UT_GENERATEGEOSHB:
            offset = dissect_ut_gn_geoshb(tvb, ut_tree, offset);
            break;
        case (guint8)UT_GENERATEGEOTSB:
            offset = dissect_ut_gn_geotsb(tvb, ut_tree, offset);
            break;
        case (guint8)UT_GNEVENTINDICATION:
            offset = dissect_ut_gn_event_indication(tvb, ut_tree, offset);
            break;
        case (guint8)UT_GNTRIGGER_RESULT:
            offset = dissect_ut_gn_trigger_result(tvb, ut_tree, offset);
            break;
        case (guint8)UT_MAPSPAT_TRIGGER_EVENT:
            offset = dissect_ut_mapspat_trigger_event(tvb, ut_tree, offset);
            break;
        case (guint8)UT_MAPSPAT_TRIGGER_EVENT_RESULT:
            offset = dissect_ut_mapspat_trigger_event_result(tvb, ut_tree, offset);
            break;
        default:
            return 0;
    } /* End of 'swith' statement */
    
    return offset;
  }
  
  return 0;
} // End of function dissect_ut_packet

/* Register the protocol with Wireshark */
void
proto_register_ut(void)
{
  module_t *ut_module;


  /* Setup list of fields 
   * NOTE Error 'Expert Info (Warn/Malformed): Trying to fetch an unsigned integer with length 8' means that hf_xxx shall be set up with 'FT_NONE, BASE_NONE' 
   */
  static hf_register_info hf[] = {
    /* UT Command */
    { &hf_command,
      {"Command", "ut", FT_UINT8, BASE_HEX, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT Initialise */
    { &hf_initialize,
      {"UtInitialise", "ut.initialize", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },  
    { &hf_initialize_result,
      {"UtInitialiseResult", "ut.initialize.result", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT ChangePosition */
    { &hf_change_position,
      {"UtChangePosition", "ut.change_position", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },  
    { &hf_change_position_result,
      {"UtChangePositionResult", "ut.change_position.result", FT_UINT8, BASE_HEX, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT ChangePseudonym */
    { &hf_change_pseudonym,
      {"UtChangePseudonym", "ut.change_pseudonym", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },  
    { &hf_change_pseudonym_result,
      {"UtChangePseudonymResult", "ut.change_pseudonym.result", FT_UINT8, BASE_HEX, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT Change Speed */
    { &hf_change_speed,
      {"UtChangeSpeed", "ut.change_speed", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT SetAccelerationControlStatus */
    { &hf_set_acceleration_control_status,
      {"SetAccelerationControlStatus", "ut.set_acceleration_control_status", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    { &hf_set_acceleration_control_status_flags,
      {"Flags", "ut.set_acceleration_control_status.flags", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    { &hf_set_acceleration_control_status_flags_B_bit,
      {"B bit", "ut.set_acceleration_control_status.flags.bbit", FT_UINT8, BASE_HEX, NULL, 0x80, NULL, HFILL}
    },
    { &hf_set_acceleration_control_status_flags_G_bit,
      {"G bit", "ut.set_acceleration_control_status.flags.gbit", FT_UINT8, BASE_HEX, NULL, 0x40, NULL, HFILL}
    },
    { &hf_set_acceleration_control_status_flags_E_bit,
      {"E bit", "ut.set_acceleration_control_status.flags.ebit", FT_UINT8, BASE_HEX, NULL, 0x20, NULL, HFILL}
    },
    { &hf_set_acceleration_control_status_flags_C_bit,
      {"C bit", "ut.set_acceleration_control_status.flags.cbit", FT_UINT8, BASE_HEX, NULL, 0x10, NULL, HFILL}
    },
    { &hf_set_acceleration_control_status_flags_A_bit,
      {"A bit", "ut.set_acceleration_control_status.flags.abit", FT_UINT8, BASE_HEX, NULL, 0x08, NULL, HFILL}
    },
    { &hf_set_acceleration_control_status_flags_CC_bit,
      {"CC bit", "ut.set_acceleration_control_status.flags.ccbit", FT_UINT8, BASE_HEX, NULL, 0x04, NULL, HFILL}
    },
    { &hf_set_acceleration_control_status_flags_L_bit,
      {"L bit", "ut.set_acceleration_control_status.flags.lbit", FT_UINT8, BASE_HEX, NULL, 0x02, NULL, HFILL}
    },
    
    /* UT SetExteriorLightsStatus */
    { &hf_set_exterior_lights_status,
      {"SetExteriorLightsStatus", "ut.hf_set_exterior_lights_status", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    { &hf_set_exterior_lights_status_flags,
      {"Flags", "ut.set_exterior_lights_status.flags", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    { &hf_set_exterior_lights_status_flags_LB_bit,
      {"LB bit", "ut.set_exterior_lights_status.flags.lbit", FT_UINT8, BASE_HEX, NULL, 0x80, NULL, HFILL}
    },
    { &hf_set_exterior_lights_status_flags_HB_bit,
      {"HB bit", "ut.set_exterior_lights_status.flags.hbbit", FT_UINT8, BASE_HEX, NULL, 0x40, NULL, HFILL}
    },
    { &hf_set_exterior_lights_status_flags_LT_bit,
      {"LT bit", "ut.set_exterior_lights_status.flags.ltbit", FT_UINT8, BASE_HEX, NULL, 0x20, NULL, HFILL}
    },
    { &hf_set_exterior_lights_status_flags_RT_bit,
      {"RT bit", "ut.set_exterior_lights_status.flags.rtbit", FT_UINT8, BASE_HEX, NULL, 0x10, NULL, HFILL}
    },
    { &hf_set_exterior_lights_status_flags_D_bit,
      {"D bit", "ut.set_exterior_lights_status.flags.dbit", FT_UINT8, BASE_HEX, NULL, 0x08, NULL, HFILL}
    },
    { &hf_set_exterior_lights_status_flags_R_bit,
      {"R bit", "ut.set_exterior_lights_status.flags.rbit", FT_UINT8, BASE_HEX, NULL, 0x04, NULL, HFILL}
    },
    { &hf_set_exterior_lights_status_flags_F_bit,
      {"LFbit", "ut.set_exterior_lights_status.flags.fbit", FT_UINT8, BASE_HEX, NULL, 0x02, NULL, HFILL}
    },
    { &hf_set_exterior_lights_status_flags_P_bit,
      {"P bit", "ut.set_exterior_lights_status.flags.pbit", FT_UINT8, BASE_HEX, NULL, 0x01, NULL, HFILL}
    },
    
    /* UT Change Heading */
    { &hf_change_heading,
      {"UtChangeHeading", "ut.change_heading", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT Set Drive Direction */
    { &hf_set_drive_direction,
      {"SetDriveDirection", "ut.set_drive_direction", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT Change YawRate */
    { &hf_change_yawrate,
      {"ChangeYawRate", "ut.change_yawrate", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT Set Dangerous Goods */
    { &hf_set_dangerous_goods,
      {"SetDangerousGoods", "ut.set_dangerous_goods", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT Set Dangerous Goods Extended */
    { &hf_set_dangerous_goods_ext,
      {"SetDangerousGoodsExtended", "ut.set_dangerous_goods_ext", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT Set Light Bar Siren */
    { &hf_set_light_bar_siren,
      {"SetLightBarSiren", "ut.hf_set_light_bar_siren", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    { &hf_set_light_bar_siren_flags,
      {"Flags", "ut.hf_set_light_bar_siren.flags", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    { &hf_set_light_bar_siren_flags_LB_bit,
      {"LB bit", "ut.hf_set_light_bar_siren.flags.lbbit", FT_UINT8, BASE_HEX, NULL, 0x80, NULL, HFILL}
    },
    { &hf_set_light_bar_siren_flags_S_bit,
      {"S bit", "ut.hf_set_light_bar_siren.flags.sbit", FT_UINT8, BASE_HEX, NULL, 0x40, NULL, HFILL}
    },
    
    /* UT CAM Event Indication */
    { &hf_cam_event_indication,
      {"CAMEventIndication", "ut.cam_event_indication", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    { &hf_cam_event_indication_length,
      {"Length", "ut.cam_event_indication_length", FT_UINT16, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_cam_event_indication_payload,
      {"Payload", "ut.cam_event_indication_payload", FT_BYTES, BASE_NONE, NULL, 0x00, NULL, HFILL}
    },
    
    /* UT SetStationType */
    { &hf_set_station_type,
      {"SetStationType", "ut.set_station_type", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT SetVehicleRole */
    { &hf_set_vehicle_role,
      {"SetVehicleRole", "ut.set_vehicle_role", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT SetEmbarkationStatus */
    { &hf_set_embarkation_status,
      {"SetEmbarkationStatus", "ut.set_embarkation_status", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT SetPtActivation */
    { &hf_set_pt_activation,
      {"SetPtActivation", "ut.set_pt_activation", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    { &hf_set_pt_activation_length,
      {"Length", "ut.set_pt_activation", FT_UINT8, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_set_pt_activation_payload,
      {"Payload", "ut.set_pt_activation", FT_BYTES, BASE_NONE, NULL, 0x00, NULL, HFILL}
    },
    
    /* UT CAM Trigger */
    { &hf_cam_trigger_result,
      {"UtCamTriggerResult", "ut.cam.trigger.result", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT DENM Trigger */
    { &hf_denm_trigger,
      {"UtDenmTrigger", "ut.denm_trigger", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    { &hf_denm_trigger_result,
      {"UtDenmTriggerResult", "ut.denm_trigger.result", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT DENM Event */
    { &hf_denm_event,
      {"UtDenmEvent", "ut.denm_event", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    { &hf_denm_event_result,
      {"UtDenmEventResult", "ut.denm_event.result", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT Terminate DENM Event */
    { &hf_terminate_denm_event,
      {"UtTerminateDenmEvent", "ut.terminate_denm_event", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    { &hf_terminate_denm_event_result,
      {"UtTerminateDenmEventResult", "ut.terminate_denm_event.result", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT Initialize */
    { &hf_hashed_id8,
      {"HashedId8", "ut.initialize.hashedid8", FT_UINT64, BASE_HEX, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT DENM Trigger */
    { &hf_denm_trigger_flags,
      {"Flags", "ut.denm_trigger.flags", FT_UINT8, BASE_HEX, NULL, 0x00, NULL, HFILL}
    },
    { &hf_denm_trigger_flags_V_bit,
      {"V bit", "ut.denm_trigger.flags.vbit", FT_UINT8, BASE_HEX, NULL, 0x80, NULL, HFILL}
    },
    { &hf_denm_trigger_flags_R_bit,
      {"R bit", "ut.denm_trigger.flags.rbit", FT_UINT8, BASE_HEX, NULL, 0x40, NULL, HFILL}
    },
    { &hf_denm_trigger_flags_K_bit,
      {"K bit", "ut.denm_trigger.flags.kbit", FT_UINT8, BASE_HEX, NULL, 0x04, NULL, HFILL}
    },
    
    /* UT DENM Event */
    { &hf_denm_event_flags,
      {"Flags", "ut.denm_event.flags", FT_UINT8, BASE_HEX, NULL, 0x00, NULL, HFILL}
    },
    { &hf_denm_event_flags_V_bit,
      {"V bit", "ut.denm_event.flags.vbit", FT_UINT8, BASE_HEX, NULL, 0x80, NULL, HFILL}
    },
    { &hf_denm_event_flags_S_bit,
      {"S bit", "ut.denm_event.flags.sbit", FT_UINT8, BASE_HEX, NULL, 0x40, NULL, HFILL}
    },
    { &hf_denm_event_flags_D_bit,
      {"D bit", "ut.denm_event.flags.dbit", FT_UINT8, BASE_HEX, NULL, 0x20, NULL, HFILL}
    },
    { &hf_denm_event_flags_T_bit,
      {"T bit", "ut.denm_event.flags.tbit", FT_UINT8, BASE_HEX, NULL, 0x10, NULL, HFILL}
    },
    { &hf_denm_event_flags_C_bit,
      {"C bit", "ut.denm_event.flags.cbit", FT_UINT8, BASE_HEX, NULL, 0x08, NULL, HFILL}
    },
    { &hf_denm_event_flags_K_bit,
      {"K bit", "ut.denm_event.flags.kbit", FT_UINT8, BASE_HEX, NULL, 0x04, NULL, HFILL}
    },
    { &hf_denm_event_flags_I_bit,
      {"I bit", "ut.denm_event.flags.ibit", FT_UINT8, BASE_HEX, NULL, 0x02, NULL, HFILL}
    },
    { &hf_denm_event_flags_X_bit,
      {"K bit", "ut.denm_event.flags.xbit", FT_UINT8, BASE_HEX, NULL, 0x01, NULL, HFILL}
    },

    /* UT DENM Event Indication */
    { &hf_denm_event_indication,
      {"DENMEventIndication", "ut.denm_event_indication", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    { &hf_denm_event_indication_length,
      {"Length", "ut.denm_event_indication_length", FT_UINT16, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_denm_event_indication_payload,
      {"Payload", "ut.denm_event_indication_payload", FT_BYTES, BASE_NONE, NULL, 0x00, NULL, HFILL}
    },
    
    /* UT GN Unicast */
    { &hf_gn_geounicast,
      {"GenerateGeoUnicast", "ut.gn.geounicast", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT GN Geobroadcast */
    { &hf_gn_geobroadcast,
      {"GenerateGeoBroadcast", "ut.gn.geobroadcast", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT GN Anycast */
    { &hf_gn_geoanycast,
      {"GenerateGeoAnycast", "ut.gn.geoanycast", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT GN Shb */
    { &hf_gn_geoshb,
      {"GenerateGeoShb", "ut.gn.geoshb", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT GN Shb */
    { &hf_gn_geotsb,
      {"GenerateGeoTsb", "ut.gn.geotsb", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT GN Event Indication */
    { &hf_gn_event_indication,
      {"GNEventIndication", "ut.gn_event_indication", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    { &hf_gn_event_indication_length,
      {"Length", "ut.gn_event_indication.length", FT_UINT16, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_gn_event_indication_payload,
      {"Payload", "ut.gn_event_indication.payload", FT_BYTES, BASE_NONE, NULL, 0x00, NULL, HFILL}
    },
    
    /* UT GN Trigger Result */
    { &hf_gntrigger_result,
      {"UtGnTriggerResult", "ut.gntrigger_result.result", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* UT MAP/SPaT Trigger Event */
    { &hf_mapspat_trigger_event,
      {"UtMapSpatTrigger ", "ut.mapspat.trigger", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },

    /* UT MAP/SPaT Trigger Result */
    { &hf_mapspat_trigger_event_result,
      {"UtMapSpatTriggerResult", "ut.mapspat.trigger_result", FT_NONE, BASE_NONE, NULL, 0x0, NULL, HFILL}
    },
    
    /* Parameters */
    { &hf_delta_latitude,
      {"Delta Latitude", "ut.change_position.delta_latitude", FT_UINT32, BASE_DEC, NULL, 0x0, NULL, HFILL}
    },
    { &hf_delta_longitude,
      {"Delta Longitude", "ut.change_position.delta_longitude", FT_UINT32, BASE_DEC, NULL, 0x0, NULL, HFILL}
    },
    { &hf_delta_elevation,
      {"Delta Elevation", "ut.change_position.delta_elevation", FT_UINT8, BASE_DEC, NULL, 0x0, NULL, HFILL}
    },
    { &hf_speed_variation,
      {"SpeedVariation", "ut.speed_variation", FT_UINT16, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_heading_direction,
      {"Direction", "ut.change_heading.direction", FT_UINT16, BASE_DEC, NULL, 0x0, NULL, HFILL}
    },
    { &hf_drive_direction,
      {"Direction", "ut.set_drive_direction.direction", FT_UINT8, BASE_DEC, VALS(ut_drive_direction_names), 0x0, NULL, HFILL}
    },
    { &hf_yawrate,
      {"Yawrate", "ut.change_yawrate.yawrate", FT_UINT8, BASE_DEC, NULL, 0x0, NULL, HFILL}
    },
    { &hf_station_type,
      {"StationType", "ut.set_station_type.station_type", FT_UINT8, BASE_DEC, VALS(ut_station_type_names), 0x0, NULL, HFILL}
    },
    { &hf_vehicle_role,
      {"VehicleRole", "ut.set_vehicle_role.vehicle_role", FT_UINT8, BASE_DEC, VALS(ut_vehicle_role_names), 0x0, NULL, HFILL}
    },
    { &hf_embarkation_status,
      {"EmbarkationStatus", "ut.set_embarkation_status.embarkation_status", FT_UINT8, BASE_DEC, VALS(ut_embarkation_status_names), 0x0, NULL, HFILL}
    },
    { &hf_dangerous_goods,
      {"DangerousGoods", "ut.set_dangerous_goods.dangerous_goods", FT_UINT8, BASE_DEC, VALS(ut_setdangerousgoods_names), 0x0, NULL, HFILL}
    },
    { &hf_dangerous_goods_ext,
      {"DangerousGoodsExtended", "ut.set_dangerous_goods_ext.dangerousgoods_ext", FT_UINT8, BASE_DEC, NULL, 0x0, NULL, HFILL}
    },   
    { &hf_station_id,
      {"StationID", "ut.denm.station_id", FT_UINT32, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_sequence_number,
      {"SequenceNumber", "ut.denm.sequence_number", FT_UINT16, BASE_DEC, NULL, 0x0, NULL, HFILL}
    },
    { &hf_detection_time,
      {"DetctionTime", "ut.denm.detection_time", FT_INT64, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_validity_duration,
      {"ValidityDuration", "ut.denm.validity_duration", FT_UINT24, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_repetition_duration,
      {"RepetitionDuration", "ut.denm.flags.repetition_duration", FT_UINT24, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_info_quality,
      {"InfoQuality", "ut.denm.flags.info_quality", FT_UINT8, BASE_HEX, NULL, 0x00, NULL, HFILL}
    },
    { &hf_cause,
      {"Cause", "ut.denm.flags.cause", FT_UINT8, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_subcause,
      {"SubCause", "ut.denm.flags.subcause", FT_UINT8, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_relevance_distance,
      {"RelevanceDistance", "ut.denm.flags.relevance_distance", FT_UINT8, BASE_DEC, VALS(ut_relevancedistance_names), 0x00, NULL, HFILL}
    },
    { &hf_relevance_traffic_direction,
      {"RelevanceTrafficDirection", "ut.denm.flags.relevance_traffic_direction", FT_UINT8, BASE_DEC, VALS(ut_relevancetrafficdirection_names), 0x00, NULL, HFILL}
    },
    { &hf_transmission_interval,
      {"TransmissionInterval", "ut.denm.flags.transmission_interval", FT_UINT16, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_repetition_interval,
      {"RepetitionInterval", "ut.denm.flags.repetition_interval", FT_UINT16, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_alacarte_length,
      {"alacarteLength", "ut.denm.flags.alacarte_length", FT_UINT8, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_alacarte,
      {"alacarte", "ut.denm.flags.alacarte", FT_BYTES, BASE_NONE, NULL, 0x00, NULL, HFILL}
    },
    { &hf_gn_address,
      {"GnAddress", "ut.gn.geoxxxcast.gnaddress", FT_BYTES, BASE_NONE, NULL, 0x00, NULL, HFILL}
    },
    { &hf_shape,
      {"Shape", "ut.gn.geoxxxcast.shape", FT_UINT8, BASE_DEC, VALS(ut_shape_names), 0x00, NULL, HFILL}
    },
    { &hf_lifetime,
      {"Lifetime", "ut.gn.geoxxxcast.lifetime", FT_UINT16, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_trafficclass,
      {"TrafficClass", "ut.gn.geoxxxcast.trafficclass", FT_UINT8, BASE_DEC,NULL, 0x00, NULL, HFILL}
    },
    { &hf_reservedgeobroadcast,
      {"Reserved", "ut.gn.geoxxxcast.reservedgeobroadcast", FT_UINT24, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_latitude,
      {"Latitude", "ut.gn.geoxxxcast.latitude", FT_UINT32, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_longitude,
      {"Longitude", "ut.gn.geoxxxcast.longitude", FT_UINT32, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_distancea,
      {"DistanceA", "ut.gn.geoxxxcast.distanceA", FT_UINT16, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_distanceb,
      {"DistanceB", "ut.gn.geoxxxcast.distanceB", FT_UINT16, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_angle,
      {"Angle", "ut.gn.geoxxxcast.angle", FT_UINT16, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_payloadlength,
      {"PayloadLength", "ut.gn.geoxxxcast.payloadlength", FT_UINT16, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_payload,
      {"Payload", "ut.gn.geoxxxcast.payload", FT_BYTES, BASE_NONE, NULL, 0x00, NULL, HFILL}
    },
    { &hf_nbhops,
      {"NbHops", "ut.gn.geotsb.nbhops", FT_UINT8, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_dstgnaddr,
      {"DstGnAddr", "ut.gn.geounicast.dstgnaddr", FT_UINT64, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    { &hf_event,
      {"Event", "ut.mapspat.trigger.event", FT_UINT8, BASE_DEC, VALS(ut_mapspat_event_names), 0x00, NULL, HFILL}
    },
    
    /* UT Initialize Result code */
    { &hf_initialize_result_code,
      {"Code", "ut.initialize.result.code", FT_UINT8, BASE_HEX, VALS(ut_result_names), 0x00, NULL, HFILL}
    },

    /* UT CAM Trigger Result code */
    { &hf_cam_trigger_result_code,
      {"Code", "ut.cam.trigger.result.code", FT_UINT8, BASE_HEX, VALS(ut_result_names), 0x00, NULL, HFILL}
    },
    
    /* UT Change Position Result code */
    { &hf_change_position_result_code,
      {"Code", "ut.change_position.result.code", FT_UINT8, BASE_HEX, VALS(ut_result_names), 0x00, NULL, HFILL}
    },
    
    /* UT Change Pseudonym Result code */
    { &hf_change_pseudonym_result_code,
      {"Code", "ut.change_pseudonym.result.code", FT_UINT8, BASE_HEX, VALS(ut_result_names), 0x00, NULL, HFILL}
    },
    
    /* UT DENM Trigger Result code */
    { &hf_denm_trigger_result_code,
      {"Code", "ut.denm_trigger.result.code", FT_UINT8, BASE_HEX, VALS(ut_result_names), 0x0, NULL, HFILL}
    },
    
    /* UT DENM Event Result code */
    { &hf_denm_event_result_code,
      {"Code", "ut.denm_event.result.code", FT_UINT8, BASE_HEX, VALS(ut_result_names), 0x0, NULL, HFILL}
    },
    
    /* UT Terminate Denm Event Result code */
    { &hf_denm_terminate_event_result_code,
      {"Code", "ut.terminate_denm_event.result.code", FT_UINT8, BASE_HEX, VALS(ut_result_names), 0x00, NULL, HFILL}
    },
    
    /* UT GN Trigger Result code */
    { &hf_gntrigger_result_code,
      {"Code", "ut.gntrigger_result.result.code", FT_UINT8, BASE_HEX, VALS(ut_result_names), 0x00, NULL, HFILL}
    },

    /* UT MAP/SPaT Trigger Result code */
    { &hf_mapspat_trigger_event_result_code,
      {"Code", "ut.mapspat_result.result.code", FT_UINT8, BASE_HEX, VALS(ut_result_names), 0x00, NULL, HFILL}
    },
    { &hf_mapspat_trigger_event_result_msgc,
      {"Message Count", "ut.mapspat_result.result.msgc", FT_UINT8, BASE_DEC, NULL, 0x00, NULL, HFILL}
    },
    
  };
  
  /* Setup protocol subtree array */
  static gint *ett[] = {
    &ett_ut,
    &ett_ut_command,
    &ett_ut_flags,
    &ett_ut_data
  };

  /* Register the protocol name and description */
  proto_ut = proto_register_protocol (
				       "ETSI ITS Upper Tester protocol",/* name       */
				       "UT",                            /* short name */
				       "ut"                             /* abbrev     */
				       );
//  new_register_dissector("ut", dissect_ut_packet, proto_ut);
  register_dissector("ut", dissect_ut_packet, proto_ut);

  /* Required function calls to register the header fields and subtrees used */
  proto_register_field_array(proto_ut, hf, array_length(hf));
  proto_register_subtree_array(ett, array_length(ett));
  register_dissector_table("ut.command", "UT command", FT_UINT8, BASE_DEC);
  
  /* Register preferences module */
  ut_module = prefs_register_protocol(proto_ut, proto_reg_handoff_ut);
}

void
proto_reg_handoff_ut(void)
{
  dissector_handle_t ut_handle;
  ut_handle = new_create_dissector_handle(dissect_ut_packet, proto_ut);
  dissector_add_uint("udp.port", UT_PROTOCOL_PORT, ut_handle);
}
